import json
import base64
import functools

from typing import Optional

from automon.log import logger
from automon.integrations.requestsWrapper import Requests

from .action_run import ActionRun
from .artifact import Artifact
from .config import SplunkSoarConfig
from .container import Container
from .vault import Vault
from .rest import Urls
from .responses import (
    AppRunResults,
    AppRunResponse,
    CancelPlaybookResponse,
    CloseContainerResponse,
    CreateContainerAttachmentResponse,
    CreateContainerResponse,
    GenericResponse,
    PlaybookRun,
    Response,
    RunPlaybookResponse,
    UpdatePlaybookResponse,
    VaultResponse
)

log = logger.logging.getLogger(__name__)
log.setLevel(logger.DEBUG)

logger.logging.getLogger('RequestsClient').setLevel(logger.DEBUG)


class SplunkSoarClient:
    def __init__(self, host: str = None,
                 user: str = None,
                 password: str = None,
                 config: SplunkSoarConfig = None):
        """Splunk SOAR Client"""

        self.config = config or SplunkSoarConfig(host=host, user=user, password=password)
        self.client = Requests(headers=self.config.headers)

        self.action_run = None
        self.app = None
        self.app_run = None
        self.asset = None
        self.containers = None
        self.playbook_run = None

    def __repr__(self) -> str:
        return f'{self.__dict__}'

    def _content(self) -> bytes:
        """get result"""
        if self.client.results:
            return self.client.results.content
        return b''

    def _content_dict(self) -> dict:
        """convert request.content to dict"""
        if self._content():
            return json.loads(self._content())
        return {}

    def _get(self, url: str) -> bool:
        """send get request"""
        return self.client.get(url=url, headers=self.client.headers)

    def _is_connected(func):
        """wrapper for connection checking"""

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if self.config.is_ready:
                if self._get(Urls.container(page_size=1)):
                    return func(self, *args, **kwargs)
            return False

        return wrapper

    def _delete(self, url: str) -> bool:
        """send get request"""
        return self.client.delete(url=url, headers=self.client.headers)

    def _post(self, url: str, data: dict) -> bool:
        """send post request"""
        return self.client.post(url=url, headers=self.client.headers, data=data)

    @_is_connected
    def close_container(self, container_id: int, **kwargs) -> Optional[CloseContainerResponse]:
        """Set container status to closed"""
        data = dict(status='closed')
        if self._post(Urls.container(identifier=container_id, **kwargs), data=json.dumps(data)):
            if self.client.results.status_code == 200:
                response = CloseContainerResponse(self._content_dict())
                log.info(f'container closed: {response}')
                return response

        log.error(msg=f'close failed. {self.client.to_dict()}')

    @_is_connected
    def cancel_playbook_run(
            self,
            playbook_run_id: int = None,
            cancel: bool = True, **kwargs) -> CancelPlaybookResponse:
        """Cancel playbook run"""
        data = dict(cancel=cancel)
        data = json.dumps(data)

        if self._post(Urls.playbook_run(identifier=playbook_run_id, **kwargs), data=data):
            if self.client.results.status_code == 200:
                response = CancelPlaybookResponse(self._content_dict())
                log.info(f'cancel playbook run: {response}')
                return response

        log.error(f'cancel failed: {playbook_run_id} {self.client.to_dict()}')

    @_is_connected
    def create_artifact(
            self,
            container_id,
            cef=None,
            cef_types=None,
            data=None,
            description=None,
            end_time=None,
            ingest_app_id=None,
            kill_chain=None,
            label=None,
            name=None,
            owner_id=None,
            run_automation=None,
            severity=None,
            source_data_identifier=None,
            start_time=None,
            tags=None,
            type=None,
            *args, **kwargs) -> Artifact:
        """Create artifact"""

        artifact = Artifact(
            dict(cef=cef,
                 cef_types=cef_types,
                 container_id=container_id,
                 data=data,
                 description=description,
                 end_time=end_time,
                 ingest_app_id=ingest_app_id,
                 kill_chain=kill_chain,
                 label=label,
                 name=name,
                 owner_id=owner_id,
                 run_automation=run_automation,
                 severity=severity,
                 source_data_identifier=source_data_identifier,
                 start_time=start_time,
                 tags=tags,
                 type=type)
        )

        if self._post(Urls.artifact(*args, **kwargs), data=artifact.to_json()):
            if self.client.results.status_code == 200:
                id = self.client.to_dict()['id']
                log.info(f'artifact created. {artifact} {self.client.to_dict()}')
                return self.get_artifact(artifact_id=id)
            else:
                existing_artifact_id = self.client.to_dict()['existing_artifact_id']
                log.info(f'artifact exists. {artifact} {self.client.to_dict()}')
                return self.get_artifact(artifact_id=existing_artifact_id)

        log.error(f'create artifact. {self.client.to_dict()}')
        return False

    @_is_connected
    def create_container(
            self,
            label,
            name,
            artifacts=None,
            asset_id=None,
            close_time=None,
            custom_fields=None,
            data=None,
            description=None,
            due_time=None,
            end_time=None,
            ingest_app_id=None,
            kill_chain=None,
            owner_id=None,
            role_id=None,
            run_automation=None,
            sensitivity=None,
            severity=None,
            source_data_identifier=None,
            start_time=None,
            open_time=None,
            status=None,
            tags=None,
            tenant_id=None,
            container_type=None,
            template_id=None,
            authorized_users=None,
            *args, **kwargs) -> Container:
        """Create container"""

        container = Container(
            dict(label=label,
                 name=name,
                 artifacts=artifacts,
                 asset_id=asset_id,
                 close_time=close_time,
                 custom_fields=custom_fields,
                 data=data,
                 description=description,
                 due_time=due_time,
                 end_time=end_time,
                 ingest_app_id=ingest_app_id,
                 kill_chain=kill_chain,
                 owner_id=owner_id,
                 role_id=role_id,
                 run_automation=run_automation,
                 sensitivity=sensitivity,
                 severity=severity,
                 source_data_identifier=source_data_identifier,
                 start_time=start_time,
                 open_time=open_time,
                 status=status,
                 tags=tags,
                 tenant_id=tenant_id,
                 container_type=container_type,
                 template_id=template_id,
                 authorized_users=authorized_users)
        )

        if self._post(Urls.container(*args, **kwargs), data=container.to_json()):
            if self.client.results.status_code == 200:
                response = CreateContainerResponse(self.client.to_dict())
                log.info(f'container created. {container} {response}')
                return response
        log.error(f'create container. {self.client.to_dict()}')
        return False

    @staticmethod
    def base64_encode(data: bytes, **kwargs) -> str:
        encode = base64.b64encode(data, **kwargs)
        decode = encode.decode()
        return decode

    @_is_connected
    def create_container_attachment(
            self,
            container_id: int,
            file_name: str,
            file_content: bytes,
            metadata: dict = None, **kwargs) -> Optional[CreateContainerAttachmentResponse]:
        """Create container attachment"""

        if metadata:
            metadata = json.dumps(metadata)

        file_content = self.base64_encode(file_content)

        data = json.dumps(dict(
            container_id=container_id,
            file_name=file_name,
            file_content=file_content,
            metadata=metadata
        ))

        if self._post(Urls.container_attachment(**kwargs), data=data):
            response = CreateContainerAttachmentResponse(self.client.to_dict())
            log.info(f'create attachment: {response}')
            return response

        log.error(f'create attachment failed.')

    @_is_connected
    def create_vault(
            self,
            file_location,
            container_id=None,
            file_name=None,
            metadata=None,
            trace=False, **kwargs) -> Vault:
        """Add vault object"""

        data = Vault(dict(
            container=container_id,
            file_location=file_location,
            file_name=file_name,
            metadata=metadata,
            trace=trace
        ))

        if self._post(Urls.vault_add(identifire=data.id, **kwargs), data=data.to_json()):
            response = Vault(self._content_dict())
            log.info(msg=f'add vault: {response}')
            return response

        log.error(msg=f'add vault failed.')

    @_is_connected
    def delete_container(self, container_id, *args, **kwargs):
        """Delete containers"""
        assert isinstance(container_id, int)

        if self._delete(Urls.container(identifier=container_id, *args, **kwargs)):
            if self.client.results.status_code == 200:
                log.info(f'container deleted: {container_id}')
                return True
        log.error(f'delete container: {container_id}. {self.client.to_dict()}')
        return False

    def is_connected(self) -> bool:
        """check if client can connect"""
        if self.config.is_ready:
            if self._get(Urls.container(page_size=1)):
                log.info(f'client connected '
                         f'{self.config.host} '
                         f'[{self.client.results.status_code}] ')
                return True

        else:
            log.warning(f'client not connected')
        return False

    @_is_connected
    def filter_vault(self, filter: str, page_size: int = None, **kwargs) -> [Vault]:
        """Filter for matching vault files"""
        matches = []

        for sublist in self.list_vault_generator(page_size=page_size, **kwargs):
            for vault in sublist:
                if filter in vault.meta.values():
                    matches.append(vault)
                elif filter in vault.names:
                    matches.append(vault)
                elif filter in vault.__dict__.values():
                    matches.append(vault)

        return matches

    @_is_connected
    def generic_delete(self, api: str, **kwargs) -> Optional[GenericResponse]:
        """Make generic delete calls"""
        if self._delete(Urls.generic(api=api, **kwargs)):
            response = GenericResponse(self._content_dict())
            log.info(f'generic delete {api}: {response}')
            return response

        log.error(f'failed generic delete {api}')

    @_is_connected
    def generic_get(self, api: str, **kwargs) -> Optional[GenericResponse]:
        """Make generic get calls"""
        if self._get(Urls.generic(api=api, **kwargs)):
            response = GenericResponse(self._content_dict())
            log.info(f'generic get {api}: {response}')
            return response

        log.error(f'failed generic get {api}')

    @_is_connected
    def generic_post(self, api: str, data: dict, **kwargs) -> Optional[GenericResponse]:
        """Make generic post calls"""
        if self._post(Urls.generic(api=api, **kwargs), data=data):
            response = GenericResponse(self._content_dict())
            log.info(f'generic post {api}: {response}')
            return response

        log.error(f'failed generic post {api}')

    @_is_connected
    def get_action_run(self, action_run_id: int = None, **kwargs) -> ActionRun:
        """Get action run"""
        if self._get(Urls.action_run(identifier=action_run_id, **kwargs)):
            action_run = ActionRun(self._content_dict())
            log.info(f'get action run: {action_run}')
            return action_run

        log.error(f'action run not found: {action_run_id}')
        return ActionRun()

    @_is_connected
    def get_artifact(self, artifact_id: int = None, **kwargs) -> Artifact:
        """Get artifact"""
        if self._get(Urls.artifact(identifier=artifact_id, **kwargs)):
            artifact = Artifact(self._content_dict())
            log.info(f'get artifact: {artifact}')
            return artifact

        log.error(f'artifact not found: {artifact_id}')
        return Artifact()

    @_is_connected
    def get_container(self, container_id: int = None, **kwargs) -> Container:
        """Get container"""
        if self._get(Urls.container(identifier=container_id, **kwargs)):
            container = Container(self._content_dict())
            log.info(f'get container: {container}')
            return container

        log.error(f'container not found: {container_id}')
        return Container()

    @_is_connected
    def get_playbook_run(self, playbook_run_id: str, **kwargs) -> Optional[PlaybookRun]:
        """Get running playbook"""
        if self._get(Urls.playbook_run(identifier=playbook_run_id, **kwargs)):
            response = PlaybookRun(self._content_dict())

            if response.status != 'failed':
                log.info(f'playbook run: {response}')
                return response

            log.error(f'playbook run failed: {response.message_to_dict}')
            return response

        log.error(f'playbook failed: {self.client.errors}')

    @_is_connected
    def get_vault(self, vault_id: int, **kwargs) -> Optional[Vault]:
        """Get vault object"""
        if self._get(Urls.vault(identifier=vault_id, **kwargs)):
            if self.client.results.status_code == 200:
                response = Vault(self._content_dict())
                log.info(msg=f'get vault: {response}')
                return response

        log.error(msg=f'get vault failed: {self.client.to_dict()}')

    @_is_connected
    def list_artifact(self, **kwargs) -> Response:
        """list artifacts"""
        if self._get(Urls.artifact(**kwargs)):
            response = Response(self._content_dict())
            log.info(f'list artifacts: {response.count}')
            return response

        return Response()

    @_is_connected
    def list_action_run(self, **kwargs) -> bool:
        """list action run"""
        if self._get(Urls.action_run(**kwargs)):
            self.containers = self._content_dict()
            return True
        return False

    @_is_connected
    def list_app(self, **kwargs) -> bool:
        """list app"""
        if self._get(Urls.app(**kwargs)):
            self.app = self._content_dict()
            return True
        return False

    @_is_connected
    def list_app_run(
            self,
            page: int = None,
            page_size: int = None, **kwargs) -> bool:
        """list app run"""
        if self._get(Urls.app_run(page=page, page_size=page_size, **kwargs)):
            self.app_run = AppRunResponse(self._content_dict())
            response = AppRunResponse(self._content_dict())
            response.data = [AppRunResults(x) for x in response.data]
            log.info(f'list app runs, page: {page} page_size: {page_size}, {response.summary()}')
            self.app_run = response
            return response
        return False

    @_is_connected
    def list_app_run_generator(
            self,
            page: int = 0,
            page_size: int = None,
            max_pages: int = None, **kwargs) -> AppRunResults or bool:
        """Generator for paging through app runs"""

        page = page

        while True:
            response = self.list_app_run(page=page, page_size=page_size, **kwargs)
            if response.data:
                app_runs = response.data
                num_pages = response.num_pages
                log.info(f'{page}/{num_pages} ({round(page / num_pages * 100, 2)}%)')

                if page >= num_pages or page >= max_pages:
                    log.info(f'list app runs finished')
                    return True

                yield app_runs
                page += 1

            elif response.data == []:
                log.info(f'{page}/{num_pages} ({round(page / num_pages * 100, 2)}%)')
                log.info(f'list app runs finished. {response}')
                return True

            elif response.data is None:
                log.error(f'list app runs failed')
                return False

            else:
                log.info(f'no app runs. {response}')
                return True

        return False

    @_is_connected
    def list_app_run_by_playbook_run(
            self,
            playbook_run: int,
            page: int = None,
            page_size: int = None, **kwargs) -> bool:
        """list app run by playbook run"""

        app_runs = []

        for app_run in self.list_app_run_generator(page=page, page_size=page_size, **kwargs):
            if app_run.playbook_run == playbook_run:
                app_runs.append(app_run)

        return app_runs

    @_is_connected
    def list_artifacts(
            self,
            page: int = None,
            page_size: int = 1000, **kwargs) -> Response:
        """list artifacts"""
        if self._get(Urls.artifact(page=page, page_size=page_size, **kwargs)):
            response = Response(self._content())
            log.info(f'list artifacts: {len(response.data)}')
            return response
        return Response()

    @_is_connected
    def list_artifact_generator(
            self,
            page: int = 0,
            page_size: int = None,
            max_pages: int = None, **kwargs) -> [Container]:
        """Generator for paging through artifacts"""

        page = page

        while True:
            response = self.list_containers(page=page, page_size=page_size, **kwargs)
            if response.data:
                containers = [Container(x) for x in response.data]
                num_pages = response.num_pages
                log.info(f'{page}/{num_pages} ({round(page / num_pages * 100, 2)}%)')

                if page >= num_pages or page >= max_pages:
                    log.info(f'list container finished')
                    return True

                yield containers
                page += 1

            elif response.data == []:
                log.info(f'{page}/{num_pages} ({round(page / num_pages * 100, 2)}%)')
                log.info(f'list container finished. {response}')
                return True

            elif response.data is None:
                log.error(f'list container failed')
                return False

            else:
                log.info(f'no containers. {response}')
                return True

        return False

    @_is_connected
    def list_asset(self, **kwargs) -> Response:
        """list asset"""
        if self._get(Urls.asset(**kwargs)):
            response = Response(self._content_dict())
            log.info(f'list assets: {len(response.data)}')
            return response
        return Response()

    @_is_connected
    def list_containers(
            self,
            page: int = None,
            page_size: int = 1000,
            *args, **kwargs) -> Response:
        """list containers"""

        url = Urls.container(page=page, page_size=page_size, *args, **kwargs)
        if self._get(url):
            response = Response(self._content_dict())
            log.info(f'list containers: {len(response.data)}')
            return response
        log.error(f'no containers')
        return Response()

    @_is_connected
    def list_containers_generator(
            self,
            page: int = 0,
            page_size: int = None,
            max_pages: int = None, **kwargs) -> [Container]:
        """Generator for paging through containers"""

        page = page
        i = 0

        while True:
            if max_pages and i > max_pages:
                break

            response = self.list_containers(
                page=page,
                page_size=page_size, **kwargs
            )

            i += 1
            if response.data:
                containers = [Container(x) for x in response.data]
                num_pages = response.num_pages
                log.info(f'container page {page}/{num_pages} ({round(page / num_pages * 100, 2)}%)')

                if page > num_pages:
                    log.info(f'list container finished')
                    break

                yield containers
                page += 1

            elif response.data == []:
                log.info(f'{page}/{num_pages} ({round(page / num_pages * 100, 2)}%)')
                log.info(f'list container finished. {response}')
                break

            elif response.data is None:
                log.error(f'list container failed')
                break

            else:
                log.info(f'no containers. {response}')
                break

        return []

    @_is_connected
    def list_cluster_node(self, **kwargs) -> Optional[dict]:
        """List cluster node"""
        if self._get(Urls.cluster_node(**kwargs)):
            cluster_node = self._content_dict()
            return cluster_node

    @_is_connected
    def list_vault(self, **kwargs) -> Optional[VaultResponse]:
        """List vault"""
        if self._get(Urls.vault(**kwargs)):
            response = VaultResponse(self._content_dict())
            log.info(msg=f'list vault: {response}')
            return response

        log.error(msg=f'list vault failed.')

    @_is_connected
    def list_vault_generator(
            self,
            page: int = 0,
            page_size: int = None,
            max_pages: int = None, **kwargs) -> [Vault]:
        """Generator for paging through vaults"""
        i = 0

        while True:
            if max_pages and i >= max_pages:
                break

            response = self.list_vault(
                page=page,
                page_size=page_size, **kwargs
            )

            i += 1
            if response.data:
                vaults = [Vault(x) for x in response.data]
                num_pages = response.num_pages
                log.info(f'vault page {page}/{num_pages} ({round(page / num_pages * 100, 2)}%)')

                if page > num_pages:
                    log.info(f'list vault finished')
                    break

                yield vaults
                page += 1

            elif response.data == []:
                log.info(f'{page}/{num_pages} ({round(page / num_pages * 100, 2)}%)')
                log.info(f'list vault finished. {response}')
                break

            elif response.data is None:
                log.error(f'list vault failed')
                break

            else:
                log.info(f'no vaults. {response}')
                break

        return []

    @_is_connected
    def update_playbook(
            self,
            playbook_id: int = None,
            active: bool = None,
            cancel_runs: bool = False,
            **kwargs) -> Optional[UpdatePlaybookResponse]:
        """Update playbook active state"""
        data = dict(
            active=active,
            cancel_runs=cancel_runs
        )
        data = json.dumps(data)
        if self._post(Urls.playbook(identifier=playbook_id, **kwargs), data=data):
            if self.client.results.status_code == 200:
                response = UpdatePlaybookResponse(self._content_dict())
                log.info(f'update playbook: {data}')
                return response

        log.error(f'update failed: {self.client.to_dict()}')

    @_is_connected
    def run_playbook(
            self,
            container_id: int,
            playbook_id: int,
            scope: str = 'all',
            run: bool = True,
            **kwargs) -> Optional[RunPlaybookResponse]:
        """Run playbook on a container"""
        data = dict(
            container_id=container_id,
            playbook_id=playbook_id,
            scope=scope,
            run=run
        )
        data = json.dumps(data)
        if self._post(Urls.playbook_run(**kwargs), data=data):
            if self.client.results.status_code == 200:
                response = RunPlaybookResponse(self._content_dict())
                log.info(f'run playbook: {data}')
                return response

        log.error(f'run failed: {self.client.to_dict()}')



import scrapy

from automon.log import Logging

log = Logging(name='ScrapyClient', level=Logging.DEBUG)


class ScrapyClient(object):

    def Selector(self, text: str):
        return scrapy.selector.Selector(text=text)

    def xpath(self, text: str, xpath: str):
        return self.Selector(text=text).xpath(xpath).get()



import logging

from .phantom_unittest import import_playbook, container, results, artifact

timestamp = '%(asctime)s'
levelname = '%(levelname)s'
modname = '%(name)s'
message = '%(message)s'
log_format = f'{timestamp}\t{levelname}\t({modname})\t{message}'

logging.basicConfig(level=logging.DEBUG, format=log_format)
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def playbook(playbook_to_import: str, container: dict = container(),
             name: str = '', callback: object = None):
    """Mock function"""
    log = logging.getLogger(f'{__name__}.{playbook.__name__}')

    # log.info(f'playbook_to_import: str, container: dict = {}, name: str, callback: object')

    log.debug(f'playbook_to_import: {playbook_to_import}')
    log.debug(f'container: {container}')
    log.debug(f'name: {name}')
    log.debug(f'callback: {callback}')

    return import_playbook(playbook_to_import)


def get_run_data(key: str, **kwargs):
    """Mock function"""
    log = logging.getLogger(f'{__name__}.{get_run_data.__name__}')

    # log.info(f'key: {key}')
    log.debug(f'key: {key}')

    return key


def condition(container: dict = container(), conditions: list = [], name: str = ''):
    """Mock function"""
    log = logging.getLogger(f'{__name__}.{condition.__name__}')

    log.debug(f'{help(condition)}')

    log.debug(f'container: {container}')
    log.debug(f'conditions: {conditions}')
    log.debug(f'name: {name}')

    return container, conditions, name


def format(container: dict = container(), template: str = '', parameters: list = [str],
           name: str = ''):
    """Mock function"""
    log = logging.getLogger(f'{__name__}.{format.__name__}')

    log.debug('container: dict = {}, template: str = '', parameters: list = [], name: str = ''')

    parameters_orig = parameters

    parameters = [
        [x.split(':')] for x in parameters
    ]

    log.debug(f'container: {container}')
    log.debug(f'template: {template}')
    log.debug(f'parameters: {parameters}')
    log.debug(f'name: {name}')

    return container, template, parameters, name


def get_format_data(name: str, **kwargs):
    """Mock function"""
    log = logging.getLogger(f'{__name__}.{get_format_data.__name__}')

    log.debug(f'name: {name}')

    return name


def collect2(container: dict = container(), datapath: list = [], action_results: object = None):
    """Mock function"""
    log = logging.getLogger(f'{__name__}.{collect2.__name__}')

    log.info('container: dict = {}, datapath: list = [], action_results: object = None')

    log.debug(f'container: {container}')
    log.debug(f'datapath: {datapath}')
    log.debug(f'action_results: {action_results}')

    return [[artifact(), artifact()]]


def act(action: str, parameters: str, assets: list, callback: object, name: str, parent_action: str = ''):
    """Mock function"""
    log = logging.getLogger(f'{__name__}.{act.__name__}')

    # log.info(f'action: str, parameters: str, assets: list, callback: object, name: str')

    log.debug(f'action: {action}')
    log.debug(f'parameters: {parameters}')
    log.debug(f'assets: {assets}')
    log.debug(f'callback: {callback}')
    log.debug(f'name: {name}')
    log.debug(f'parent_action: {parent_action}')

    return action, parameters, assets, callback, name, parent_action


def save_run_data(key: str, value: str, auto: bool):
    """Mock function"""
    log = logging.getLogger(f'{__name__}.{save_run_data.__name__}')

    log.debug(f'key: {key}')
    log.debug(f'value: {value}')
    log.debug(f'auto: {auto}')

    return key, value, auto


def debug(object: str):
    return log.debug(f'{object}')


def error(object: str):
    return log.error(f'{object}')



from automon.log import logger

from .config import InstagramConfig

log = logger.logging.getLogger(__name__)
log.setLevel(logger.INFO)


class InstagramClient(object):

    def __init__(self, login: str = None, password: str = None, config: InstagramConfig = None):
        """Instagram Client"""
        self.config = config or InstagramConfig(login=login, password=password)
        self.login = self.config.login

    def _isAuthenticated(self):
        return

    def isAuthenticated(self):
        return

    def __repr__(self):
        return f'{self.__dict__}'



from automon.log import logger

log = logger.logging.getLogger(__name__)
log.setLevel(logger.INFO)


def set_window_size(width: int = 1920, height: int = 1080, device_type: str = None) -> (int, int):
    """set browser resolution"""

    if device_type == 'pixel3':
        width = 1080
        height = 2160

    if device_type == 'web-small' or device_type == '800x600':
        width = 800
        height = 600

    if device_type == 'web-small-2' or device_type == '1024x768':
        width = 1024
        height = 768

    if device_type == 'web-small-3' or device_type == '1280x960':
        width = 1280
        height = 960

    if device_type == 'web-small-4' or device_type == '1280x1024':
        width = 1280
        height = 1024

    if device_type == 'web' or device_type == '1920x1080':
        width = 1920
        height = 1080

    if device_type == 'web-2' or device_type == '1600x1200':
        width = 1600
        height = 1200

    if device_type == 'web-3' or device_type == '1920x1200':
        width = 1920
        height = 1200

    if device_type == 'web-large' or device_type == '2560x1400':
        width = 2560
        height = 1400

    if device_type == 'web-long' or device_type == '1920x3080':
        width = 1920
        height = 3080

    if not width and not height:
        width = 1920
        height = 1080

    log.debug(f'{int(width)}, {int(height)}')

    return int(width), int(height)



import datetime
import logging
import automon
import tracemalloc

import pandas as pd
import numpy as np

from automon.log import logger
from automon.helpers.sleeper import Sleeper
from automon.integrations.facebook import FacebookGroups
from automon.integrations.google.sheets import GoogleSheetsClient

logging.getLogger('google_auth_httplib2').setLevel(logging.ERROR)
logging.getLogger('googleapiclient.discovery').setLevel(logging.ERROR)
logging.getLogger('googleapiclient.discovery_cache').setLevel(logging.ERROR)
logging.getLogger('urllib3.connectionpool').setLevel(logging.ERROR)
logging.getLogger('selenium.webdriver.common.service').setLevel(logging.ERROR)
logging.getLogger('selenium.webdriver.remote.remote_connection').setLevel(logging.ERROR)
logging.getLogger('selenium.webdriver.common.selenium_manager').setLevel(logging.ERROR)

logging.getLogger('automon.integrations.seleniumWrapper.browser').setLevel(logging.DEBUG)
logging.getLogger('automon.integrations.seleniumWrapper.config_webdriver_chrome').setLevel(logging.DEBUG)
logging.getLogger('automon.integrations.facebook.groups').setLevel(logging.DEBUG)
logging.getLogger('automon.integrations.requestsWrapper.client').setLevel(logging.INFO)

tracemalloc.start()

log = logger.logging.getLogger(__name__)
log.setLevel(logger.DEBUG)

SHEET_NAME = 'Automated Count DO NOT EDIT'
SHEET_NAME_INGEST = 'URLS TO INGEST'

sheets_client = GoogleSheetsClient(
    worksheet=SHEET_NAME,
)


def get_facebook_info(url: str):
    if not url:
        return {}

    group = FacebookGroups(
        url=url_cleaner(url=url)
    )
    # group.start(headless=False)
    group.start(headless=True)
    group.get_about()

    return group.to_dict()


def url_cleaner(url: str):
    if not url:
        return
    if url[-1] == '/':
        url = url[:-1]
    return url


def merge_urls():
    sheets_client.get(
        ranges='AUG23 AUDIT!A:Z',
        fields="sheets/data/rowData/values/hyperlink",
    )

    data = sheets_client.response['sheets'][0]['data'][0]['rowData']
    # expand nested data
    links = []
    for x in data:
        if x:
            links.append(
                x['values'][0]['hyperlink']
            )

    df_Shelley = pd.DataFrame(data=links, columns=['url'])

    sheets_client.get()
    sheets_client.get_values(
        range=f'{SHEET_NAME}!A:Z'
    )

    sheet_values = sheets_client.values
    sheet_columns = sheet_values[0]
    sheet_data = sheet_values[1:]

    df = pd.DataFrame(data=sheet_data, columns=sheet_columns)
    df = df.dropna(subset=['url'])

    # merge both lists or urls
    df = pd.merge(df, df_Shelley, how='outer', on='url')
    df = df.drop_duplicates(subset=['url'], keep='first')
    return df


def batch_processing(sheet_index: int, df: pd.DataFrame):
    df_results = df['url'].dropna().apply(
        lambda url: get_facebook_info(url=url)
    )
    df_results = pd.DataFrame(df_results.tolist())

    df = df.reset_index()

    todays_date = datetime.datetime.now().date()
    monthly = f'{todays_date.year}-{todays_date.month}'

    # create columns
    df[f'url'] = df_results['url']
    df[f'{monthly}'] = df_results['members_count']
    df[f'last_updated'] = monthly
    df[f'title'] = df_results['title']
    df[f'content_unavailable'] = df_results['content_unavailable']
    df[f'creation_date'] = df_results['creation_date']
    df[f'creation_date_timestamp'] = df_results['creation_date_timestamp']
    df[f'history'] = df_results['history']
    df[f'members_count'] = df_results['members_count']
    df[f'posts_monthly_count'] = df_results['posts_monthly_count']
    df[f'posts_today_count'] = df_results['posts_today_count']
    df[f'privacy'] = df_results['privacy']
    df[f'visible'] = df_results['visible']

    # set dtype to Int32
    df[f'{monthly}'] = df[f'{monthly}'].astype('Int32')
    df[f'creation_date_timestamp'] = df[f'creation_date_timestamp'].astype('Int32')
    df[f'members_count'] = df[f'members_count'].astype('Int32')
    df[f'posts_monthly_count'] = df[f'posts_monthly_count'].astype('Int32')
    df[f'posts_today_count'] = df[f'posts_today_count'].astype('Int32')

    # order columns
    columns = [
        'url',
        'title',
        'creation_date',
        'creation_date_timestamp',
        'history',
        'privacy',
        'visible',
        'content_unavailable',
        'last_updated',
        'posts_monthly_count',
        'posts_today_count',
        'members_count',
    ]

    sheet_index_df = df['index'].loc[0]
    assert sheet_index == sheet_index_df
    df = df.drop('index', axis=1)

    # add all other dates
    df_columns = df.columns.tolist()
    columns.extend(
        [x for x in df_columns if x not in columns]
    )

    # finally add today's date
    if f'{monthly}' not in columns:
        columns.append(
            f'{monthly}',
        )

    df = df.loc[:, columns]
    df = df.fillna(np.nan).replace([np.nan], [None])

    update_columns = sheets_client.update(
        range=f'{SHEET_NAME}!A1:Z',
        values=[columns],
    )

    update = sheets_client.update(
        range=f'{SHEET_NAME}!A{sheet_index_df}:Z',
        values=[x for x in df.values.tolist()]
    )

    log.info(f'{sheet_index_df}: {[x for x in df.values.tolist()]}')

    return df


def memory_profiler():
    snapshot = tracemalloc.take_snapshot()
    top_stats = snapshot.statistics("lineno")

    df_memory_profile = pd.DataFrame([
        dict(size_B=stat.size, count=stat.count, file=stat.traceback._frames[0][0],
             file_line=stat.traceback._frames[0][1]) for stat in top_stats
    ])
    df_memory_profile.sort_values(by='size_B', ascending=False)
    df_memory_profile['size_KB'] = df_memory_profile['size_B'].apply(
        lambda B: round(B / 1024)
    )
    df_memory_profile['size_MB'] = df_memory_profile['size_KB'].apply(
        lambda KB: round(KB / 1024)
    )
    cols = df_memory_profile.columns.tolist()
    cols.sort()
    df_memory_profile = df_memory_profile.loc[:, cols]

    log.debug(
        f"total memory used: {df_memory_profile['size_MB'].sum()} MB; "
        f'most memory used: '
        f'{df_memory_profile.iloc[0].to_dict()}'
    )

    return df_memory_profile


def main():
    if not sheets_client.authenticate():
        return

    # merge urls from audit sheet
    # df_audit = merge_urls()
    # df_audit = df_audit.fillna(np.nan).replace([np.nan], [None])
    # rows = []
    # rows.append(df_audit.columns.tolist())
    # rows.extend([x for x in df_audit.values.tolist()])
    # update = sheets_client.update(
    #     range=f'{SHEET_NAME}!A:Z',
    #     values=rows
    # )

    # start processing
    sheets_client.get_values(
        range=f'{SHEET_NAME}!A:Z'
    )

    sheet_values = sheets_client.values
    sheet_columns = sheet_values[0]
    sheet_data = sheet_values[1:]

    df = pd.DataFrame(data=sheet_data, columns=sheet_columns)
    df = df.dropna(subset=['url'])
    # set df index to match google sheet index numbering
    df.index = df.index + 2

    # drop duplicates
    df_duplicates_to_delete = df[
        (df.url.duplicated(keep='first'))
    ]

    for duplicate in df_duplicates_to_delete.iterrows():
        duplicate_index, duplicate_row = duplicate

        # clear row in sheet
        range = f'{SHEET_NAME}!{duplicate_index}:{duplicate_index}'
        result = sheets_client.clear(range=range)
        # max 60/min
        Sleeper.seconds(seconds=1)
        log.info(result)
        df = df.drop(duplicate_index)

    # ingest urls from SHEET_NAME_INGEST
    sheets_client.get_values(
        range=f'{SHEET_NAME_INGEST}!A:Z'
    )
    ingest_sheet_values = sheets_client.values
    if ingest_sheet_values:
        ingest_sheet_values = [x[0] if x else [] for x in ingest_sheet_values]
        ingest_sheet_values = [url_cleaner(x) for x in ingest_sheet_values]
    df_ingest_sheet_values = pd.DataFrame(ingest_sheet_values)
    df_ingest_sheet_values.index = df_ingest_sheet_values.index + 1

    for ingest_data in df_ingest_sheet_values.iterrows():
        ingest_index, ingest_url = ingest_data
        ingest_url = ingest_url[0]

        if ingest_url not in df['url'].values:
            ingest_series = pd.Series({'url': ingest_url}).to_frame().T
            index_add_url = df.index[-1] + 1
            df.loc[index_add_url] = {'url': ingest_url}

            df.loc[index_add_url] = df.loc[index_add_url].fillna(np.nan).replace([np.nan], [None])

            values = [[x for x in df.loc[index_add_url].values.tolist()]]

            update = sheets_client.update(
                range=f'{SHEET_NAME}!A{index_add_url}:Z',
                values=values
            )

            log.info(
                f'{index_add_url}: {values}'
            )

        # clear url from ingest sheet
        range = f'{SHEET_NAME_INGEST}!{ingest_index}:{ingest_index}'
        clear = sheets_client.clear(range=range)
        log.info(f'{clear}')

    # start updating
    for data in df.iterrows():
        data_index, data_row = data

        df_batch = df.loc[data_index:data_index]

        # skip if last_updated is the current month
        todays_date = datetime.datetime.now().date()
        last_updated = f'{todays_date.year}-{todays_date.month}'
        if df_batch['last_updated'].iloc[0] == last_updated:
            # log.debug(f'skipping {data_index}, {data_row.to_dict()}')
            continue

        batch_result = batch_processing(sheet_index=data_index, df=df_batch)

        try:
            batch_result = batch_processing(sheet_index=data_index, df=df_batch)
            df_memory = memory_profiler()
        except Exception as e:
            df_memory = memory_profiler()
            log.error(f'{e}')

    pass


if __name__ == '__main__':
    main()



from automon.log import Logging

from .person import Person

log = Logging(name='GooglePeopleResults', level=Logging.DEBUG)


class ConnectionsResults:
    connections: list
    nextPageToken: str
    nextSyncToken: str
    totalPeople: int
    totalItems: int

    def __init__(self, result: dict):
        """people.connections.list results

        {
          "connections": [
            {
              object (Person)
            }
          ],
          "nextPageToken": string,
          "nextSyncToken": string,
          "totalPeople": integer,
          "totalItems": integer
        }
        """

        self.connections = None
        self.nextPageToken = None
        self.nextSyncToken = None
        self.totalItems = None
        self.totalPeople = None

        self.__dict__.update(result)

        log.debug(msg=f'{self}')

    def __repr__(self):
        return f'totalPeople: {self.totalPeople}, totalItems: {self.totalItems}, contacts: {len(self.contacts)}'

    def __eq__(self, other):
        if isinstance(other, ConnectionsResults):
            return self.__dict__ == other.__dict__
        return NotImplemented

    @property
    def contacts(self) -> [Person]:
        return [Person(x) for x in self.connections]



import os
import json
import base64
import datetime
import tempfile
import selenium
import selenium.webdriver

from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from urllib.parse import urlparse
from bs4 import BeautifulSoup

from automon.log import logger
from automon.helpers.dates import Dates
from automon.helpers.sleeper import Sleeper
from automon.helpers.sanitation import Sanitation

from .config import SeleniumConfig
from .user_agents import SeleniumUserAgentBuilder

log = logger.logging.getLogger(__name__)
log.setLevel(logger.DEBUG)


class SeleniumBrowser(object):
    config: SeleniumConfig
    webdriver: selenium.webdriver
    status: int

    def __init__(self, config: SeleniumConfig = None):
        """A selenium wrapper"""

        self._config = config or SeleniumConfig()

    def __repr__(self):
        try:
            return str(dict(
                webdriver=self.webdriver.name or None,
                window_size=self.window_size,
            ))
        except Exception as error:
            pass

        return f'{__class__}'

    @property
    def by(self) -> By:
        """Set of supported locator strategies"""
        return selenium.webdriver.common.by.By()

    @property
    def config(self):
        return self._config

    def cookie_file_to_dict(self, file: str = 'cookies.txt'):
        log.debug(f'{file}')
        with open(file, 'r') as file:
            return json.loads(file.read())

    @property
    def _current_url(self):
        try:
            self.webdriver.current_url
            return self.webdriver.current_url
        except Exception as error:
            return

    @property
    def webdriver(self):
        return self.config.webdriver

    @property
    def get_log(self) -> list:
        """Gets the log for a given log type"""
        logs = []
        for log_type in self.webdriver.log_types:
            logs.append(
                {
                    log_type: self.webdriver.get_log(log_type)
                }
            )

        return logs

    @property
    def keys(self):
        """Set of special keys codes"""
        return selenium.webdriver.common.keys.Keys

    def refresh(self):
        self.webdriver.refresh()
        log.info(f'{True}')

    @property
    def url(self):
        return self.current_url

    @property
    def user_agent(self):
        try:
            return self.webdriver.execute_script("return navigator.userAgent")
        except:
            return None

    @property
    def current_url(self):
        if self.webdriver:
            log.debug(self._current_url)
            if self._current_url == 'data:,':
                return ''
            return self._current_url

        log.info(None)
        return ''

    @property
    def window_size(self):
        return self.config.webdriver_wrapper.window_size

    def _screenshot_name(self, prefix=None):
        """Generate a unique filename"""

        title = self.webdriver.title
        url = self._current_url
        hostname = urlparse(url).hostname

        hostname_ = Sanitation.ascii_numeric_only(hostname)
        title_ = Sanitation.ascii_numeric_only(title)
        timestamp = Dates.filename_timestamp()

        if prefix:
            prefix = Sanitation.safe_string(prefix)
            return f'{prefix}_{hostname_}_{title_}_{timestamp}.png'

        return f'{hostname_}_{title_}_{timestamp}.png'

    def action_click(self, xpath: str, note: str = None) -> str or False:
        """perform mouse command"""
        try:
            click = self.find_element(value=xpath, by=self.by.XPATH)
            click.click()
            if note:
                log.debug(str(dict(
                    note=note,
                    xpath=xpath,
                )))
            else:
                log.debug(str(dict(
                    xpath=xpath,
                )))
            return click

        except Exception as error:
            message, session, stacktrace = self.error_parsing(error)
            log.error(str(dict(
                url=self.url,
                xpath=xpath,
                message=message,
                session=session,
                stacktrace=stacktrace,
            )))
        return False

    def action_type(self, key: str or Keys, secret: bool = True):
        """perform keyboard command"""
        try:
            actions = selenium.webdriver.common.action_chains.ActionChains(
                self.webdriver)
            actions.send_keys(key)
            actions.perform()

            if secret:
                key = f'*' * len(key)

            log.debug(str(dict(
                send_keys=key,
            )))
            return True

        except Exception as error:
            message, session, stacktrace = self.error_parsing(error)
            log.error(str(dict(
                url=self.url,
                send_keys=key,
                message=message,
                session=session,
                stacktrace=stacktrace,
            )))
        return False

    def add_cookie(self, cookie_dict: dict) -> bool:
        result = self.webdriver.add_cookie(cookie_dict=cookie_dict)

        if result is None:
            log.debug(f'{cookie_dict}')
            return True

        log.error(f'{cookie_dict}')
        return False

    def add_cookie_from_file(self, file: str) -> bool:
        """add cookies from file"""
        if os.path.exists(file):
            self.add_cookies_from_list(
                self.cookie_file_to_dict(file=file)
            )
            return True

        log.error(f'{file}')
        return False

    def add_cookies_from_list(self, cookies_list: list):
        for cookie in cookies_list:
            self.add_cookie(cookie_dict=cookie)

        log.debug(f'{True}')
        return True

    def add_cookie_from_current_url(self):
        log.info(f'{self.url}')
        return self.add_cookie_from_url(self.url)

    def add_cookie_from_url(self, url: str):
        """add cookies from matching hostname"""
        cookie_file = self._url_filename(url=url)

        if os.path.exists(cookie_file):
            log.info(f'{cookie_file}')
            return self.add_cookie_from_file(file=cookie_file)

        log.error(f'{cookie_file}')

    def add_cookie_from_base64(self, base64_str: str):
        if base64_str:
            self.add_cookies_from_list(
                json.loads(base64.b64decode(base64_str))
            )
            log.debug(f'{True}')
            return True

        log.error(f'{base64_str}')
        return False

    def delete_all_cookies(self) -> None:
        result = self.webdriver.delete_all_cookies()
        log.info(f'{True}')
        return result

    def _url_filename(self, url: str):
        parsed = self.urlparse(url)
        hostname = parsed.hostname
        cookie_file = f'cookies-{hostname}.txt'
        log.info(f'{cookie_file}')
        return cookie_file

    def get_cookie(self, name: str) -> dict:
        result = self.webdriver.get_cookie(name=name)
        log.info(f'{result}')
        return result

    def get_cookies(self) -> [dict]:
        result = self.webdriver.get_cookies()
        log.debug(f'{True}')
        return result

    def get_cookies_base64(self) -> base64:
        result = self.get_cookies()
        log.debug(f'{True}')
        return base64.b64encode(
            json.dumps(result).encode()
        ).decode()

    def get_cookies_json(self) -> json.dumps:
        cookies = self.get_cookies()
        log.debug(f'{True}')
        return json.dumps(cookies)

    def get_cookies_summary(self):
        result = self.get_cookies()
        summary = {}
        if result:
            for cookie in result:
                cookie = dict(cookie)
                domain = cookie.get('domain')
                name = cookie.get('name')
                expiry = cookie.get('expiry')

                if domain in summary.keys():
                    summary[domain].append(cookie)
                else:
                    summary[domain] = [cookie]

        log.debug(f'{summary}')
        return summary

    def close(self):
        """close browser"""
        log.info(f'closed')
        self.webdriver.close()

    @staticmethod
    def error_parsing(error) -> tuple:
        error_parsed = f'{error}'.splitlines()
        error_parsed = [f'{x}'.strip() for x in error_parsed]
        message = error_parsed[0]
        session = error_parsed[1]
        stacktrace = error_parsed[2:]
        stacktrace = ' '.join(stacktrace)

        return message, session, stacktrace

    def find_element(
            self,
            value: str,
            by: By.ID = By.ID,
            **kwargs):
        """find element"""
        element = self.webdriver.find_element(value=value, by=by, **kwargs)
        log.info(str(dict(
            url=self.url,
            text=element.text,
            value=value,
        )))
        return element

    def find_xpath(self, value: str, by: By = By.XPATH, **kwargs):
        """find xpath"""
        xpath = self.find_element(value=value, by=by, **kwargs)
        log.info(str(dict(
            url=self.url,
            text=xpath.text,
            value=value,
        )))
        return xpath

    def get(self, url: str, **kwargs) -> bool:
        """get url"""
        try:
            if self.webdriver.get(url, **kwargs) is None:
                log.info(str(dict(
                    url=url,
                    current_url=self._current_url,
                    kwargs=kwargs
                )))
            return True
        except Exception as error:
            log.error(str(dict(
                error=error,
            )))

        return False

    def get_page(self, *args, **kwargs):
        """alias to get"""
        return self.get(*args, **kwargs)

    def get_page_source(self) -> str:
        """get page source"""
        return self.webdriver.page_source

    def get_page_source_beautifulsoup(self, markdup: str = None, features: str = 'lxml') -> BeautifulSoup:
        """read page source with beautifulsoup"""
        if not markdup:
            markdup = self.get_page_source()
        return BeautifulSoup(markup=markdup, features=features)

    def get_random_user_agent(self, filter: list or str = None, case_sensitive: bool = False) -> str:
        return SeleniumUserAgentBuilder().get_random_agent(filter=filter, case_sensitive=case_sensitive)

    def get_screenshot_as_base64(self, **kwargs):
        """screenshot as base64"""
        screenshot = self.webdriver.get_screenshot_as_base64(**kwargs)
        log.debug(f'{round(len(screenshot) / 1024)} KB')

        return screenshot

    def get_screenshot_as_file(
            self,
            filename: str = None,
            prefix: str = None,
            folder: str = None,
            **kwargs
    ) -> bool:
        return self.save_screenshot(
            self,
            filename=filename,
            prefix=prefix,
            folder=folder,
            **kwargs
        )

    def get_screenshot_as_png(self, **kwargs):
        """screenshot as png"""
        screenshot = self.webdriver.get_screenshot_as_png(**kwargs)
        log.debug(f'{round(len(screenshot) / 1024)} KB')

        return screenshot

    def is_running(self) -> bool:
        """browser is running"""
        if self.webdriver:
            log.info(f'{True}')
            return True
        log.error(f'{False}')
        return False

    def urlparse(self, url: str):
        parsed = urlparse(url=url)
        log.debug(f'{parsed}')
        return parsed

    def quit(self) -> bool:
        """gracefully quit browser"""
        try:
            self.webdriver.close()
            self.webdriver.quit()
            self.webdriver.stop_client()
        except Exception as error:
            message, session, stacktrace = self.error_parsing(error)
            log.error(str(dict(
                message=message,
                session=session,
                stacktrace=stacktrace,
            )))
            return False
        return True

    def run(self):
        """run browser"""
        return self.config.run()

    def save_cookies_for_current_url(self):
        filename = self._url_filename(url=self.url)
        log.info(f'{filename}')
        return self.save_cookies_to_file(file=filename)

    def save_cookies_to_file(self, file: str):
        with open(file, 'w') as cookies:
            cookies.write(
                self.get_cookies_json()
            )

        if os.path.exists(file):
            log.info(f'{os.path.abspath(file)} ({os.stat(file).st_size} B)')
            return True

        log.error(f'{file}')
        return False

    def save_screenshot(
            self,
            filename: str = None,
            prefix: str = None,
            folder: str = None,
            **kwargs) -> bool:
        """save screenshot to file"""

        if not filename:
            filename = self._screenshot_name(prefix)

        if not folder:
            path = os.path.abspath(tempfile.gettempdir())
        else:
            path = os.path.abspath(folder)

        if not os.path.exists(path):
            os.makedirs(path)

        save = os.path.join(path, filename)

        if self.webdriver.save_screenshot(save, **kwargs):
            log.info(f'Saving screenshot to: {save} ({round(os.stat(save).st_size / 1024)} KB)')
            return True

        return False

    def set_window_size(self, width=1920, height=1080, device_type=None) -> bool:
        """set browser resolution"""

        try:
            self.config.webdriver_wrapper.set_window_size(width=width, height=height,
                                                          device_type=device_type)
        except Exception as error:
            message, session, stacktrace = self.error_parsing(error)
            log.error(str(dict(
                message=message,
                session=session,
                stacktrace=stacktrace,
            )))
            return False
        return True

    def set_window_position(self, x: int = 0, y: int = 0):
        """set browser position"""
        result = self.webdriver.set_window_position(x, y)
        log.info(f'{result}')
        return result

    def start(self):
        """alias to run"""
        return self.run()

    def wait_for(
            self,
            value: str or list,
            by: By = By.XPATH,
            retries: int = 5,
            fail_on_error: bool = True,
            **kwargs) -> str or False:
        """wait for something"""
        if not retries:
            retries = 5

        retry = 0
        while True:
            try:
                if isinstance(value, list):
                    values = value
                    for value in values:
                        try:
                            self.find_element(
                                by=by,
                                value=value,
                                **kwargs)
                            log.debug(str(dict(
                                by=by,
                                url=self.url,
                                value=value,
                            )))
                            return value
                        except:
                            log.error(str(dict(
                                by=by,
                                url=self.url,
                                value=value,
                            )))
                else:
                    self.find_element(
                        by=by,
                        value=value,
                        **kwargs)
                    log.debug(str(dict(
                        by=by,
                        url=self.url,
                        value=value,
                    )))
                    return value
            except Exception as error:
                log.error(str(dict(
                    by=by,
                    url=self.url,
                    value=value,
                    error=error,
                )))
                Sleeper.seconds(0.2)

            retry += 1

            log.error(str(dict(
                url=self.url,
                retry=f'{retry}/{retries}',
            )))

            if retry == retries:
                break

        if fail_on_error:
            raise Exception(str(dict(
                by=by,
                url=self.url,
                value=value,
            )))

        return False

    def wait_for_element(
            self,
            element: str or list,
            retries: int = None,
            fail_on_error: bool = True,
            **kwargs
    ) -> str or False:
        """wait for an element"""
        return self.wait_for(
            value=element,
            by=self.by.ID,
            retries=retries,
            fail_on_error=fail_on_error,
            **kwargs
        )

    def wait_for_xpath(
            self,
            xpath: str or list,
            retries: int = None,
            fail_on_error: bool = True,
            **kwargs
    ) -> str or False:
        """wait for an xpath"""
        return self.wait_for(
            value=xpath,
            by=self.by.XPATH,
            retries=retries,
            fail_on_error=fail_on_error,
            **kwargs
        )



import unittest

from automon.integrations.elasticsearch.config import ElasticsearchConfig, SnapshotBot, JVMBot
from automon.integrations.elasticsearch.client import ElasticsearchClient
from automon.integrations.elasticsearch.cleanup import Cleanup
from automon.integrations.elasticsearch.metrics import Metric, MetricTimestamp, Cluster


class Elasticsearch(unittest.TestCase):
    e = ElasticsearchClient()

    def test_create_document(self):
        es = ElasticsearchClient()

        from datetime import datetime

        doc = {
            'timestamp': datetime.now(),
            'yohji': 'yamamoto'

        }

        if self.e.isConnected():
            self.assertTrue(es.create_document(doc=doc))
            self.assertTrue(es.search_summary())

    def test_ElasticsearchClient(self):
        e = ElasticsearchClient()

        if e.isConnected():
            self.assertTrue(e)
            self.assertTrue(ElasticsearchClient)
            self.assertTrue(e.ping())
            self.assertTrue(e.get_indices())
            self.assertFalse(e.delete_index(None))
            self.assertFalse(e.search_indices(None))
        else:
            self.assertFalse(e.ping())
            self.assertFalse(e.delete_index(None))
            self.assertFalse(e.search_indices(None))
            self.assertFalse(e.get_indices())

    def test_Cleanup(self):
        if self.e.isConnected():
            self.assertTrue(Cleanup().get_indices())
            self.assertFalse(Cleanup().search_indices(None))
            # self.assertFalse(Cleanup().delete_indices(None))
        else:
            self.assertFalse(Cleanup().get_indices())
            self.assertFalse(Cleanup().search_indices(None))
            # self.assertFalse(Cleanup().delete_indices(f''))

    def test_ElasticsearchConfig(self):
        self.assertTrue(ElasticsearchConfig())
        self.assertEqual(ElasticsearchConfig(), ElasticsearchConfig())

    def test_SnapshotBot(self):
        self.assertTrue(SnapshotBot())

    def test_JVMBot(self):
        self.assertTrue(JVMBot())

    def test_Cluster(self):
        self.assertTrue(Cluster)

    def test_MetricTimestamp(self):
        # metric = Metric()
        self.assertTrue(MetricTimestamp)
        # self.assertTrue(MetricTimestamp(metric)._time_now())

    def test_Metric(self):
        node = None
        jvm = {
            "timestamp": 1571330469200,
            "uptime_in_millis": 12251148809,
            "mem": {
                "heap_used_in_bytes": 27551308288,
                "heap_used_percent": 82,
                "heap_committed_in_bytes": 33216266240,
                "heap_max_in_bytes": 33216266240,
                "non_heap_used_in_bytes": 162665664,
                "non_heap_committed_in_bytes": 204083200,
                "pools": {
                    "young": {
                        "used_in_bytes": 130099216,
                        "max_in_bytes": 558432256,
                        "peak_used_in_bytes": 558432256,
                        "peak_max_in_bytes": 558432256
                    },
                    "survivor": {
                        "used_in_bytes": 1132144,
                        "max_in_bytes": 69730304,
                        "peak_used_in_bytes": 69730304,
                        "peak_max_in_bytes": 69730304
                    },
                    "old": {
                        "used_in_bytes": 27420076928,
                        "max_in_bytes": 32588103680,
                        "peak_used_in_bytes": 30895029472,
                        "peak_max_in_bytes": 32588103680
                    }
                }
            },
            "threading": {
                "count": 141,
                "peak_count": 223
            },
            "gc": {
                "collectors": {
                    "young": {
                        "collection_count": 533686,
                        "collection_time_in_millis": 37099480
                    },
                    "old": {
                        "collection_count": 75872,
                        "collection_time_in_millis": 9588732
                    }
                }
            },
            "buffer_pools": {
                "mapped": {
                    "count": 7988,
                    "used_in_bytes": 3715149748692,
                    "total_capacity_in_bytes": 3715149748692
                },
                "direct": {
                    "count": 10146,
                    "used_in_bytes": 166764364,
                    "total_capacity_in_bytes": 166764363
                }
            },
            "classes": {
                "current_loaded_count": 17808,
                "total_loaded_count": 18183,
                "total_unloaded_count": 375
            }
        }
        self.assertTrue(Metric)
        self.assertRaises(TypeError, Metric)


if __name__ == '__main__':
    unittest.main()



from automon.log import Logging
from automon.integrations.datascience import Pandas


class Other:
    def __init__(self, csv):
        self._log = Logging(name=Other.__name__, level=Logging.DEBUG)

        self.csv = csv
        self.df = OtherCSV(csv).df

    def __repr__(self):
        return f'{self.csv}'

    def __eq__(self, other):
        if self.df.equals(other.df):
            self._log.debug(f'same {other}')
            return True
        self._log.debug(f'different {other}')
        return False


class OtherCSV:
    def __init__(self, csv):
        self._log = Logging(name=OtherCSV.__name__, level=Logging.DEBUG)

        self.csv = csv
        self.df = Pandas().read_csv(csv)

    def __repr__(self):
        return f'{self.df}'

    def __eq__(self, other):
        if isinstance(other, OtherCSV):
            if self.df == other.df:
                self._log.debug(f'same {other}')
                return True
            self._log.debug(f'different {other}')
        return False



import time
import random

from automon.log import logger

log = logger.logging.getLogger(__name__)
log.setLevel(logger.DEBUG)


class Sleeper:

    @staticmethod
    def seconds(seconds: int) -> time.sleep:
        """Sleep for this many seconds"""

        sleep = seconds
        log.debug(f'{sleep}')
        return time.sleep(sleep)

    @staticmethod
    def minute(minutes: int = 1) -> time.sleep:
        """Sleep for a minute"""

        sleep = round(minutes * 60)
        log.debug(f'{sleep}')
        return time.sleep(sleep)

    @staticmethod
    def within_a_minute(sleep: int = None):
        """Sleep for a random minute"""

        sleep = sleep if isinstance(sleep, int) else \
            random.choice(range(1, 1 * 60))
        log.debug(f'{sleep}')
        return time.sleep(sleep)

    @staticmethod
    def minutes(minutes: int):
        """Sleep for this many minutes"""

        sleep = minutes * 60
        log.debug(f'{sleep}')
        return time.sleep(sleep)

    @staticmethod
    def hour(hour: int = 1):
        """At some time within an hour, this will run"""

        sleep = hour if not hour else random.choice(
            range(1, hour * 60 * 60))
        log.debug(f'{sleep}')
        return time.sleep(sleep)

    @staticmethod
    def hours(hours):
        """Sleep for this many hours"""

        sleep = hours * 60 * 60
        log.debug(f'{sleep}')
        return time.sleep(sleep)

    @staticmethod
    def day(hours: int = 24):
        """At some time within 24 hours, this will run"""

        sleep = hours if not hours else random.choice(
            range(1, hours * 60 * 60))
        log.debug(f'{sleep}')
        return time.sleep(sleep)

    @staticmethod
    def daily(hours: int = 24):
        """Sleep for one day"""

        sleep = hours if not hours else hours * 60 * 60
        log.debug(f'{sleep}')
        return time.sleep(sleep)

    @staticmethod
    def time_range(seconds: int):
        """Sleep for a random range"""

        sleep = seconds if not seconds else random.choice(
            range(1, seconds))
        log.debug(f'{sleep}')
        return time.sleep(sleep)



import functools

from automon.log import logger
from automon.integrations.seleniumWrapper.browser import SeleniumBrowser
from automon.integrations.seleniumWrapper.config_webdriver_chrome import ChromeWrapper

from automon.helpers.sleeper import Sleeper
# from automon.integrations.minioWrapper import MinioClient

from .config import InstagramConfig
from .urls import Urls
from .xpaths import XPaths

log = logger.logging.getLogger(__name__)
log.setLevel(logger.DEBUG)


class InstagramBrowserClient:
    login: str
    password: str
    config: InstagramConfig
    browser: SeleniumBrowser

    def __init__(self,
                 login: str = None,
                 password: str = None,
                 config: InstagramConfig = None,
                 headless: bool = True):
        """Instagram Browser Client"""
        self.config = config or InstagramConfig(login=login, password=password)
        self.browser = SeleniumBrowser()
        self.browser.config.webdriver_wrapper = ChromeWrapper()

        self.useragent = self.browser.get_random_user_agent()

        if headless:
            self.browser.config.webdriver_wrapper.in_headless().set_user_agent(self.useragent)
        else:
            self.browser.config.webdriver_wrapper.set_user_agent(self.useragent)

    def __repr__(self):
        return f'{self.__dict__}'

    def _is_running(func):
        @functools.wraps(func)
        def wrapped(self, *args, **kwargs):
            if self.is_running():
                return func(self, *args, **kwargs)
            return False

        return wrapped

    def _is_authenticated(func):
        @functools.wraps(func)
        def wrapped(self, *args, **kwargs):
            if self.is_authenticated():
                return func(self, *args, **kwargs)
            return False

        return wrapped

    def get_page(self, account: str):
        """ Get page
        """
        log.debug(f'[get_page] getting {account}')

        page = f'https://instagram.com/{account}'
        browser = self.authenticated_browser
        return browser.get(page)

    def get_stories(self, account: str):
        """ Retrieve story
        """
        story = f'https://www.instagram.com/stories/{account}/'
        num_of_stories = 0

        log.debug(f'[get_stories] {story}')

        browser = self.authenticated_browser
        browser.get(story)
        browser.browser.save_screenshot_to_minio(bucket_name='screenshots',
                                                 prefix='instagram/' + account)

        if 'Page Not Found' in browser.browser.title:
            log.debug(f'[get_stories] no stories for {account}')
            return num_of_stories

        Sleeper.seconds(2)

        while True:
            try:
                self._next_story(browser)

                title = browser.browser.title
                if title == 'Instagram':
                    log.debug(
                        ('[get_stories] {} end of stories'.format(account)))
                    raise Exception
                num_of_stories += 1
                browser.save_screenshot_to_minio(bucket_name='screenshots',
                                                 prefix='instagram/' + account)
                Sleeper.seconds(1)
                browser.save_screenshot_to_minio(bucket_name='screenshots',
                                                 prefix='instagram/' + account)
            except Exception as error:
                # TODO: disable browser proxy when done
                log.debug(f'[get_stories] done: {account}, {error}')
                return num_of_stories

    def _next_story(self, authenticated_browser):
        """ Click next story button
        """

        xpaths = [
            '//*[@id="react-root"]/section/div/div/section/div[2]/div[1]/div/div/div[2]/div/div/button',
            '//*[@id="react-root"]/section/div/div/section/div[2]/button[2]'
        ]

        found_btn = False
        for xpath in xpaths:
            try:
                browser = authenticated_browser
                button = browser.browser.find_element_by_xpath(xpath)
                found_btn = True
                log.debug('[next_story] next story')
                return button.click()
            except Exception as error:
                log.error(f'{error}')

        if not found_btn:
            # no more stories. exit
            log.debug('[_next_story] no more stories')
            raise Exception

    def remove_not_now(self):
        """check for "save your login info" dialogue"""
        not_now = self.browser.wait_for_xpath(
            self.xpaths.save_info_not_now_div,
            fail_on_error=False
        )
        if not_now:
            self.browser.action_type(self.browser.keys.TAB)
            self.browser.action_type(self.browser.keys.TAB)
            self.browser.action_type(self.browser.keys.ENTER)
            # self.browser.action_click(not_now, 'dont save login info')

    def remove_notifications_not_now(self):
        """check for "notifications" dialogue"""
        notifications_not_now = self.browser.wait_for_xpath(
            self.xpaths.turn_on_notifications_not_now,
            fail_on_error=False
        )
        if notifications_not_now:
            self.browser.action_click(notifications_not_now, 'no notifications')

    def run_stories(self, limit=None):
        """Run
        """

        log.debug('[login] {}'.format(self.login))

        self.authenticated_browser = self.authenticate()

        # if self.authenticated_browser:
        #
        #     count = 0
        #     if limit:
        #
        #         for account in self.following:
        #             hevlog.logging.debug(
        #                 '[runrun] [{}] {} session: {}'.format(self.authenticated_browser.browser.name,
        #                                                       self.authenticated_browser.browser.title,
        #                                                       self.authenticated_browser.browser.session_id))
        #
        #             num_of_stories = get_stories(self.authenticated_browser, account)
        #
        #             hevlog.logging.info('[{}] {} stories'.format(account, num_of_stories))
        #
        #             count = count + 1
        #             if count == limit:
        #                 return
        #
        #     Sleeper.hour('instagram')
        #     self.run_stories()

    @_is_running
    def authenticate(self):
        """Authenticate to Instagram
        """

        self.browser.get(self.urls.login_page)

        # user
        login_user = self.browser.wait_for_xpath(self.xpaths.login_user)
        self.browser.action_click(login_user, 'user')
        self.browser.action_type(self.login)

        # password
        login_pass = self.browser.wait_for_xpath(self.xpaths.login_pass)
        self.browser.action_click(login_pass, 'login')
        self.browser.action_type(self.config.password)
        self.browser.action_type(self.browser.keys.ENTER)

        self.remove_notifications_not_now()
        self.remove_not_now()

        if self.is_authenticated():
            log.info(f'{True}')
            return True

        log.error(f'{False}')
        return False

    @_is_running
    @_is_authenticated
    def get_followers(self, account: str):
        url = self.urls.followers(account)
        self.browser.get(url)

    @_is_running
    def is_authenticated(self):
        try:
            if self.urls.domain not in self.browser.url:
                self.browser.get(self.urls.domain)
            self.remove_notifications_not_now()
            self.remove_not_now()
            profile_picture = self.browser.wait_for_xpath(self.xpaths.profile_picture)
            if profile_picture:
                log.info(f'{True}')
                return True
        except Exception as error:
            log.error(f'{error}')
        return False

    def is_running(self) -> bool:
        if self.config.is_configured:
            if self.browser.is_running():
                log.info(f'{True}')
                return True
        log.error(f'{False}')
        return False

    @property
    def login(self) -> str:
        return self.config.login

    @property
    def urls(self):
        return Urls()

    @property
    def xpaths(self):
        return XPaths()
